/* ANLDP 2.0
 *
 * William McCune
 * Mathematics and Computer Science Division
 * Argonne National Laboratory
 * Argonne, IL  60439
 * U.S.A.
 *
 * E-mail: mccune@mcs.anl.gov
 * Web:    http://www.mcs.anl.gov/~mccune
 *         http://www.mcs.anl.gov/AR/mace
 */

#define VERSION "2.0"
#define VDATE "May 2001"

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <unistd.h>  /* for getopt */

#ifdef TP_NAMES  /* for calls to getpwuid() and gethostname() */
#  include <sys/types.h>
#  include <pwd.h>
#  include <unistd.h>
#endif

/******** Types of exit ********/

#define MACE_ABEND_EXIT         11
#define UNSATISFIABLE_EXIT      12
#define MACE_MAX_SECONDS_EXIT   13
#define MACE_MAX_MEM_EXIT       14
#define MAX_MODELS_EXIT         15
#define ALL_MODELS_EXIT         16
#define MACE_SIGINT_EXIT        17
#define MACE_SEGV_EXIT          18
#define MACE_INPUT_ERROR_EXIT   19

/************* END OF ALL GLOBAL CONSTANT DEFINITIONS ****************/

#include "Opts.h"
#include "Stats.h"
#include "Avail.h"
#include "Clock.h"
#include "Miscellany.h"

#include "Dp.h"

int First_order;        /* used by dp.c */
int Domain_size = 0;    /* used by stats.c */
int Init_wall_seconds;  /* used by stats.c */
int Models = 0;         /* used with dp.c, miscellany.c */

/* I didn't want ANLDP to depend on any of the otter code in the
 * parent directory, so I copied the few routines that it needs
 * and put them here.
 */

/*************
 *
 *   username()
 *
 *************/

char *username(void)
{
#ifdef TP_NAMES
  /* cuserid is not on some machines;
   * getlogin doesn't always work if user is not logged in;
   * getlogin broken on Linux.  Following seems to be ok.
   */
  struct passwd *p;
  p = getpwuid(getuid());
  return(p ? p->pw_name : "???");
#else
  return("???");
#endif
}  /* username */

/*************
 *
 *   hostname()
 *
 *************/

char *hostname(void)
{
  static char host[64];
#ifdef TP_NAMES
  if (gethostname(host, 64) != 0)
    strcpy(host, "???");
#else
  strcpy(host, "???");
#endif
  return(host);
}  /* hostname */

/*************
 *
 *     wall_seconds()
 *
 *************/

long wall_seconds(void)
{
  /* This is ANSI, and it seems to work for many OS. */
  time_t i;

  i = time((time_t *) NULL);
  return((long) i);
}  /* wall_seconds */

/*************
 *
 *   char *get_time() - get a string representation of current date and time
 *
 *************/

char *get_time(void)
{
  time_t i;
  i = time((time_t *) NULL);
  return(asctime(localtime(&i)));
}  /* get_time */

/*************
 *
 *    long run_time() - Return run time in milliseconds.
 *
 *    This is used instead of the normal clock routines in case
 *    progam is complied with TP_NO_CLOCKS.
 *
 *************/

long run_time(void)
{
#ifdef TP_RUSAGE
  struct rusage r;
  long sec, usec;

  getrusage(RUSAGE_SELF, &r);
  sec = r.ru_utime.tv_sec;
  usec = r.ru_utime.tv_usec;

  return((sec * 1000) + (usec / 1000));
#else
  long ticks;
  long sec;

  ticks = clock();
  sec = ((double) ticks / CLOCKS_PER_SEC) * 1000;
  return(sec);
#endif
}  /* run_time */

/*************
 *
 *    long system_time() - Return system time in milliseconds.
 *
 *************/

long system_time(void)
{
#ifdef TP_RUSAGE
  struct rusage r;
  long sec, usec;

  getrusage(RUSAGE_SELF, &r);
  sec = r.ru_stime.tv_sec;
  usec = r.ru_stime.tv_usec;

  return((sec * 1000) + (usec / 1000));
#else
  return(0);
#endif
}  /* system_time */

/* a few stubs that are referred (but not used by anldp) to in dp.c */

void print_model(FILE *fp) {}
void print_model_portable(FILE *fp) {}
void print_model_ivy(FILE *fp) {}
void p_unit(int u) {}

/*************
 *
 *    void MACE_print_banner(argc, argv)
 *
 *************/

static void MACE_print_banner(int argc, char **argv)
{
    int i;

    printf("----- ANLDP %s, %s -----\n", VERSION, VDATE);
    printf("The process was started by %s on %s,\n%s",
	   username(), hostname(), get_time());

    printf("The command was \"");
    for(i = 0; i < argc; i++)
        printf("%s%s", argv[i], (i < argc-1 ? " " : ""));
    printf("\".\n");
    
}  /* MACE_print_banner */

/*************
 *
 *   usage_message()
 *
 *************/

static void usage_message(void)
{
    printf("\nANLDP %s -- Search for finite models.\n\n", VERSION);
    printf("\nInput clauses are taken from standard input.\n");
    printf("\ncommand-line options:\n");

    printf("   -p      | Print all models as they are found.\n");

    printf("   -m n    | Stop when the n-th model is found.\n");
    printf("   -t n    | Stop after n seconds.\n");
    printf("   -k n    | Tp_allocate at most n K bytes.\n");

    printf("   -s      | Perform unit subsumption. (Always done on input.)\n");
    printf("   -h      | Print this message.\n");
    printf("\n");
}  /* usage_message */

/*************
 *
 *   process_command_line_args()
 *
 *************/

static int process_command_line_args(int argc, char **argv)
{
    extern char *optarg;

    int c, n;
    int error = 0;

    while ((c = getopt(argc, argv, "spPIhcm:t:k:n:d:x:")) != EOF) {
	switch (c) {
	  case 's':
	    MACE_change_flag(stderr, "subsume", 1);
	    break;
	  case 'p':
	    MACE_change_flag(stderr, "print_models", 1);
	    break;
	  case 'm':
	    n = atoi(optarg);
	    if (MACE_change_parm(stderr, "max_models", n) == -1)
		error++;
	    break;
	  case 't':
	    n = atoi(optarg);
	    if (MACE_change_parm(stderr, "max_seconds", n) == -1)
		error++;
	    break;
	  case 'k':
	    n = atoi(optarg);
	    if (MACE_change_parm(stderr, "max_mem", n) == -1)
		error++;
	    break;
	  case 'h':
	  case '?':
	  default:
	    error = 1;
	    break;
	    }
	}

    return(error == 0);

}  /* process_command_line_args */

/*************
 *
 *   MACE_sig_handler()
 *
 *************/

#ifdef TP_SIGNAL
#include <signal.h>

static void MACE_sig_handler(int condition)
{
  if (condition == SIGSEGV) {

    char message[] =
      "\n"
      "+----------------------------------------------------------+\n"
      "| SEGMENTATION FAULT!!  This is probably caused by a bug   |\n"
      "| in MACE.  Please send copy of the input file to          |\n"
      "| otter@mcs.anl.gov, let us know what version of MACE you  |\n"
      "| are using, and send any other info that might be useful. |\n"
      "+----------------------------------------------------------+\n\n";

    fprintf(stderr, "%s\007", message);

    exit_with_message(MACE_SEGV_EXIT, 1);
  }
  else if (condition == SIGINT) {
    exit_with_message(MACE_SIGINT_EXIT, 1);
  }
  else {
    char s[100];
    sprintf(s, "MACE_sig_handler, cannot handle signal %d.\n", condition);
    MACE_abend(s);
  }
}  /* MACE_sig_handler */

#endif  /* TP_SIGNAL */

/*************
 *
 *    main
 *
 *************/

int main(int argc, char **argv)
{
  int rc;

#ifdef TP_SIGNAL
  signal(SIGINT, MACE_sig_handler);
  signal(SIGSEGV, MACE_sig_handler);
#endif

  MACE_print_banner(argc, argv);
  init_dp();  /* This is MACE's initialization. */
  Init_wall_seconds = wall_seconds();
  rc = process_command_line_args(argc, argv);
  if (!rc) {
    usage_message();
    exit(0);
  }

  /*************************************************************************/

  rc = read_all_clauses(stdin);
  rc = more_setup();

  MACE_CLOCK_START(DECIDE_TIME);
  rc = dp_prover();
  MACE_CLOCK_STOP(DECIDE_TIME);

  exit_with_message(rc, 1);

  exit(99);  /* This is to shut up the compiler. */

}  /* main */
